import numpy as np
from .utility import Utility
from struct import pack, unpack, calcsize

# FileHandler nuevo nombre para la clase? filehandler.py nuevo nombre para el archivo?


class RawFile:
    """
    Class for reading and writing operations on a file
    """

    def __init__(self, name, mode):
        """Open file with name and mode"""
        self.file = open(name, mode)
        # recordar depurar los *nibble si no los usamos
        self.wnibble = None  # nibble pending to write
        self.rnibble = None  # nibble pending to read
        self.queue = []  # other data pending to write

    def __enter__(self):
        """Enter the runtime context related to this object."""
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Exit the runtime context related to this object."""
        self.close()

    def write(self, fmt, *args):
        """Pack and write args with format fmt"""
        if fmt != "n":
            if self.wnibble is None:
                fmt = (
                    "!" + fmt
                )  # big-endian. handle binary data in a platform-independent way
                args = [a.encode("utf-8") if isinstance(a, str) else a for a in args]
                self.file.write(pack(fmt, *args))
            else:
                self.queue.append((fmt, args))
        else:
            raise ValueError("fmt = n. Not implemented.")

    def read(self, fmt):
        """Read data with format fmt and unpack"""
        if fmt != "n":
            fmt = "!" + fmt
            size = calcsize(fmt)
            data = self.file.read(size)
            udata = unpack(fmt, data)
            return (
                [u.decode("utf-8") if isinstance(u, bytes) else u for u in udata]
                if len(udata) > 1
                else udata[0]
            )
        else:
            raise ValueError("fmt = n. Not implemented.")

    def tell(self):
        """Return the current file position"""
        return self.file.tell()

    def seek(self, offset, whence=0):
        """Move the file pointer to the specified position"""
        return self.file.seek(offset, whence)

    def close(self):
        """Close the file"""
        if self.wnibble is not None:
            self.write("n", 0)
        self.file.close()

    # Transferred methods from RawFileHandler
    def write_header(self, header_format, *args):
        self.write(header_format, *args)

    def write_vector(self, x, x_norm_0, v_format):
        """Write a sparse vector as a list of pairs (pos, value)"""
        self.write("B", int(x_norm_0))
        position_format = "B" if len(x) <= 256 else "H"
        if x_norm_0 > 0:
            for position, value in enumerate(x):
                if value != 0:
                    self.write(position_format, position)
                    self.write("f", float(Utility.truncate(value, v_format)))

    def read_vector(self, a_cols):
        """
        Read vector from 'file'
        First read the l0 norm,
        Then read an sparse vector as a list of pairs (pos, value)
        This function is complementary to write_vector function
        """
        x_norm_0 = self.read("B")
        x = np.zeros(a_cols)

        pos_format = "B" if x.shape[0] <= 256 else "H"
        for _ in range(x_norm_0):
            pos = self.read(pos_format)
            value = self.read("f")
            x[pos] = value

        return x



import sys
from .aqmp import ImageCompressor

def main():
    if len(sys.argv) < 2:
        print("Usage: python compressor.py <image_file> --parameters")
        sys.exit(1)

    image_file = sys.argv[1]
    parameters = sys.argv[2:]

    # Default parameters for ImageCompressor
    min_sparcity = 0.1
    min_n = 8
    max_n = 32
    a_cols = 64
    max_error = 0.01
    wavelet_election = "db1"
    shuffle_dictionary = False
    v_format_precision = "f"
    apply_deflate = False

    # Parse parameters
    for param in parameters:
        key, value = param.split('=')
        if key == "min_sparcity":
            min_sparcity = float(value)
        elif key == "min_n":
            min_n = int(value)
        elif key == "max_n":
            max_n = int(value)
        elif key == "a_cols":
            a_cols = int(value)
        elif key == "max_error":
            max_error = float(value)
        elif key == "wavelet_election":
            wavelet_election = value
        elif key == "shuffle_dictionary":
            shuffle_dictionary = value.lower() == "true"
        elif key == "v_format_precision":
            v_format_precision = value
        elif key == "apply_deflate":
            apply_deflate = value.lower() == "true"

    compressor = ImageCompressor(
        min_sparcity,
        min_n,
        max_n,
        a_cols,
        max_error,
        wavelet_election,
        shuffle_dictionary,
        v_format_precision,
        apply_deflate,
    )

    if image_file.endswith(".png"):
        output_file = image_file.replace(".png", ".fif")
        compressor.encode(image_file, output_file)
    elif image_file.endswith(".fif"):
        output_file = image_file.replace(".fif", ".png")
        compressor.decode(image_file, output_file)
    else:
        print("Unsupported file format. Please use .png for encoding or .fif for decoding.")
        sys.exit(1)

if __name__ == "__main__":
    main()


import numpy as np
from math import cos, sin, pi, sqrt, log
import pywt

class BasisFunctions:
    def phi(self, x):
        return 1 if 0 <= x < 1 else 0

    def psi(self, x):
        if 0 <= x < 0.5:
            return 1
        elif 0.5 <= x < 1:
            return -1
        return 0

    def h(self, i, N):
        if i == 0:
            return self.phi

        n, k = [(n, k) for n in range(int(log(N, 2))) for k in range(2 ** n)][i - 1]
        return lambda x: 2 ** (n / 2.0) * self.psi(2 ** n * x - k)

    def v(self, h, N):
        return [h(i / float(N)) for i in range(N)]

    def Haar1_qt(self, rows, cols):
        """Haar wavelet transform matrix"""
        return np.array([self.v(self.h(i, cols), rows) for i in range(cols)]).T

    def DCT_II_f(self, k, N):
        return lambda x: np.cos(pi * (x + 0.5) * k / N)

    def w(self, k, N):
        c = sqrt(2) ** np.sign(k)
        return [c * self.DCT_II_f(k, N)(i / float(N)) for i in range(N)]

    def DCT1_qt(self, rows, cols):
        """DCT transform matrix"""
        return np.array([self.w(k, rows) for k in range(cols)]).T

    def DCT1_Haar1_qt(self, rows, cols):
        """Combined DCT and Haar wavelet transform matrix"""
        dct_matrix = self.DCT1_qt(rows, cols // 2)
        haar_matrix = self.Haar1_qt(rows, cols // 2)
        return np.concatenate((dct_matrix, haar_matrix), axis=1)

    def W(self, k1, k2, n, N):
        def ro(t):
            return [[cos(t), -sin(t)], [sin(t), cos(t)]]

        def theta(n, N):
            return pi * n / (2.0 * N)

        def g(k1, k2, N1, N2, v):
            return self.DCT_II_f(k1, N1)(v[0]) * self.DCT_II_f(k2, N2)(v[1])

        def W_elem(i, j):
            return g(k1, k2, 8, 8, np.dot(ro(theta(n, N)), [[i / 8.0], [j / 8.0]]))

        return [[W_elem(i, j) for j in range(8)] for i in range(8)]


####################################################################################
# Nuevas bases que utilizan el las transformadas wavelets de pywt en lugar de Haar #
####################################################################################

    def wavelet_transform_qt(self, rows, cols, wavelet='db1'):
        """
        Generate a matrix using the specified wavelet from the PyWavelets library.
        :param rows: Number of rows in the output matrix.
        :param cols: Number of columns in the output matrix.
        :param wavelet: Wavelet type to use. Default is 'db1' (Daubechies wavelet).
        :return: Wavelet transform matrix.
        """
        wavelet = pywt.Wavelet(wavelet)
        wavelet_basis = pywt.wavedec(np.eye(rows, cols), wavelet, level=int(log(rows, 2)))
        
        # Flatten and format the wavelet basis into a matrix similar to Haar/DCT
        wavelet_matrix = np.array(wavelet_basis[0]) 
        return wavelet_matrix

    def shuffle_columns(self, matrix):
        """Shuffles the columns of the given matrix."""
        shuffled_indices = np.random.permutation(matrix.shape[1])
        return matrix[:, shuffled_indices]

    def DCT1_wavelet_qt(self, rows, cols, wavelet='db1', shuffle=False):
        """Combined DCT and alternative wavelet transform matrix with optional column shuffling."""
        dct_matrix = self.DCT1_qt(rows, cols // 2)  
        wavelet_matrix = self.wavelet_transform_qt(rows, cols // 2, wavelet)
        
        # Hacemos un diccionario híbrido con dos transformadas pero si luego
        # en el OMP hacemos un cut en las columnas tomamos solamente la primera transformada
        # por eso quizás lo mejor sea hacer un shuffle antes
        
        combined_matrix = np.concatenate((dct_matrix, wavelet_matrix), axis=1)  

        if shuffle:
            combined_matrix = self.shuffle_columns(combined_matrix)

        return combined_matrix



import numpy as np
from sklearn.linear_model import OrthogonalMatchingPursuit
from .utility import Utility
from .basis import BasisFunctions
from anytree import Node, RenderTree, PreOrderIter

# omphandler.py nuevo nombre para el archivo?


class OMPHandler:
    """
    Manage encoding and decoding using Orthogonal Matching Pursuit (OMP) and Adaptive Quadtrees (AQ).
    Functions to visualize the tree structure obtained with AQ.
    """

    def __init__(self, min_n, max_n, a_cols, min_sparcity, max_error):
        self.min_n = min_n
        self.max_n = max_n
        self.a_cols = a_cols
        self.min_sparcity = min_sparcity
        self.max_error = max_error
        self.omp_dict = {}
        self.basis_functions = BasisFunctions()
        self.subdivision_tree = None

    def initialize_dictionary(self):
        """Initialize dictionary for OMP"""
        n_aux = self.min_n
        while n_aux <= self.max_n:
            A = self.basis_functions.DCT1_Haar1_qt(n_aux * n_aux, self.a_cols)
            self.omp_dict[n_aux] = A
            # print(f"A.shape: {A.shape}")
            n_aux *= 2

    def initialize_dictionary2(self, wavelet_election="db1", shuffle=False):
        """Initialize dictionary for OMP"""
        n_aux = self.min_n
        while n_aux <= self.max_n:
            A = self.basis_functions.DCT1_wavelet_qt(
                n_aux * n_aux, self.a_cols, wavelet=wavelet_election, shuffle=shuffle
            )
            self.omp_dict[n_aux] = A
            n_aux *= 2

    def omp_encode(self, x_list, image_data, max_error, block_size, k):
        """Process channel of image using Matching Pursuit and track subdivisions."""
        channel_processed_blocks = 0
        self.subdivision_tree = Node("root")

        for i in range(image_data.shape[0] // block_size):
            for j in range(image_data.shape[1] // block_size):
                channel_processed_blocks, x_list = self.omp_encode_recursive(
                    block_size,
                    i * block_size,
                    j * block_size,
                    k,
                    image_data,
                    max_error,
                    x_list,
                    channel_processed_blocks,
                    parent=self.subdivision_tree,
                )
        return channel_processed_blocks, x_list

    def omp_encode_recursive(
        self,
        block_size,
        from_dim0,
        from_dim1,
        k,
        image_data,
        max_error,
        x_list,
        channel_processed_blocks,
        parent,
    ):
        """Recursive OMP code with tree-based subdivision tracking."""
        sub_image_data = Utility.sub_image(image_data, block_size, from_dim0, from_dim1)
        sub_image_data = sub_image_data.flatten()
        dict_ = self.omp_dict.get(block_size)

        # if dict_.shape[1] > sub_image_data.size:
        #     dict_ = dict_[:, :sub_image_data.size]

        # n_nonzero_coefs = int(dict_.shape[1])
        # print("hyperparameter n_nonzero_coefs:", n_nonzero_coefs)
        omp = OrthogonalMatchingPursuit(
            tol=self.max_error,
            fit_intercept=False,
            normalize=False # normalize=False stops the warning message
        )  
        omp.fit(dict_, sub_image_data)
        coefs = omp.coef_
        norm_0_coefs = np.linalg.norm(coefs, 0)  # n_nonzero_coefs_ de sklearn
        # print("norm_0_coefs using linalg:", norm_0_coefs)
        # print("n_nonzero_coefs_ with scikit:", omp.n_nonzero_coefs_)
        # print("coefs:", coefs, coefs.shape)
        # assert int(norm_0_coefs) == omp.n_nonzero_coefs_

        current_node = Node(
            f"Block_{block_size}_{from_dim0}_{from_dim1}", parent=parent
        )

        min_sparcity = Utility.min_sparcity(self.min_sparcity, block_size)
        if norm_0_coefs > min_sparcity and block_size > self.min_n:
            for x_init, y_init in [
                (x, y)
                for x in [0, int(block_size / 2)]
                for y in [0, int(block_size / 2)]
            ]:
                channel_processed_blocks, x_list = self.omp_encode_recursive(
                    int(block_size / 2),
                    from_dim0 + x_init,
                    from_dim1 + y_init,
                    k,
                    image_data,
                    max_error,
                    x_list,
                    channel_processed_blocks,
                    parent=current_node,
                )
        else:
            channel_processed_blocks += 1
            x_list.append((block_size, from_dim0, from_dim1, k, coefs))

        return channel_processed_blocks, x_list

    def omp_decode(self, file, image_data, v_format_precision, processed_blocks):
        """OMP decoder for the entire channel"""
        for _ in range(processed_blocks):
            i = file.read("H")
            j = file.read("H")
            k = file.read("B")
            n = file.read("B")
            # print("i,j,k,n", i,j,k,n)
            A = self.omp_dict[n]
            coefs = np.array(file.read_vector(self.a_cols))
            output_vector = np.dot(A, coefs)
            for elem in output_vector:
                elem = Utility.truncate(elem, v_format_precision)
            image_data[i : i + n, j : j + n, k] = output_vector.reshape((n, n))
        return image_data

    #################################################
    # New functions to visualize the tree structure #
    #################################################

    def print_subdivision_tree(self):
        """Helper function to print the subdivision tree."""
        for pre, _, node in RenderTree(self.subdivision_tree):
            print(f"{pre}{node.name}")

    def summarize_subdivisions(self):
        """Summarizes the tree structure: number of subdivisions, leaf nodes, and tree depth."""
        total_nodes = len([node for node in PreOrderIter(self.subdivision_tree)])
        subdivided_nodes = len(
            [
                node
                for node in PreOrderIter(self.subdivision_tree)
                if len(node.children) > 0
            ]
        )
        leaf_nodes = total_nodes - subdivided_nodes
        max_depth = max([node.depth for node in PreOrderIter(self.subdivision_tree)])

        print(f"Total subimages: {total_nodes}")
        print(f"Subimages that performed subdivisions: {subdivided_nodes}")
        print(f"Subimages that did not subdivide (leaf nodes): {leaf_nodes}")
        print(f"Maximum depth of subdivisions: {max_depth}")



import numpy as np


class Utility:
    """
    Utility class for image processing and numerical operations
    """

    @staticmethod
    def truncate(value, format_spec):
        """Truncate the value to the specified format."""
        try:
            return float(f"{value:{format_spec}}")
        except ValueError as error:
            raise ValueError(
                f"Invalid format code '{format_spec}' for value '{value}'"
            ) from error

    @staticmethod
    def quantize(array, v_format):
        """Truncate elements of 'array' using v_format"""
        for elem in array:
            elem = Utility.truncate(elem, v_format)
        return array

    @staticmethod
    def mode_to_bpp(mode):
        """Convert image mode to bits per pixel."""
        if mode == "L":
            return 8
        elif mode == "RGB":
            return 24
        elif mode == "RGBA":
            return 32
        elif mode == "YCbCr":
            return 24
        else:
            raise ValueError(f"Unsupported image mode: {mode}")

    @staticmethod
    def min_sparcity(min_sparcity_rate, n):
        """Calculate minimum sparsity based on max_error and block size n."""
        return min_sparcity_rate * n**2

    @staticmethod
    def sub_image(image_data, n, i, j, k=None):
        """Extract a sub-image from the larger image array."""
        h0, h1, w0, w1 = i, i + n, j, j + n
        if k is None:
            return image_data[h0:h1, w0:w1]
        return image_data[h0:h1, w0:w1, k]

    @staticmethod
    def set_sub_image(sub_img, image_data, n, i, j, k):
        """Place a sub-image back into the larger image array."""
        h0, h1, w0, w1 = i, i + n, j, j + n
        image_data[h0:h1, w0:w1, k] = sub_img

    @staticmethod
    def ycbcr_to_rgb(image_data):
        """
        Reemplazar esta función por la función de la libreria Pillow:
        image = Image.fromarray(image_data, 'YCbCr')
        """
        Y = image_data[:, :, 0]
        Cb = image_data[:, :, 1] - 128
        Cr = image_data[:, :, 2] - 128

        R = Y + 1.402 * Cr
        G = Y - 0.344136 * Cb - 0.714136 * Cr
        B = Y + 1.772 * Cb

        image_data[:, :, 0] = np.clip(R, 0, 255)
        image_data[:, :, 1] = np.clip(G, 0, 255)
        image_data[:, :, 2] = np.clip(B, 0, 255)

        return image_data



import numpy as np
from .rawfile import RawFile
from .omp import OMPHandler
from .utility import Utility
from PIL import Image
import zlib  # to apply DEFLATE

# Codec nuevo nombre para la clase?


class ImageCompressor:
    """
    Class for compressing and decompressing images
    """

    def __init__(
        self,
        min_sparcity,
        min_n,
        max_n,
        a_cols,
        max_error,
        wavelet_election="db1",
        shuffle_dictionary=False,
        v_format_precision="f",
        apply_deflate=False,
    ):
        # Algorithm Parameters
        self.min_n = min_n
        self.max_n = max_n
        self.a_cols = a_cols
        self.min_sparcity = min_sparcity
        self.max_error = max_error
        self.v_format_precision = v_format_precision
        self.apply_deflate = apply_deflate

        # Other File Format parameters (No son más argumentos de la clase)
        self.fif_version = 2
        self.magic_number = b"FIF"  # Ensure this is a bytes object
        self.header_format = "3sBiiBBBBB"

        self.omp_handler = OMPHandler(
            self.min_n, self.max_n, self.a_cols, self.min_sparcity, self.max_error
        )
        self.omp_handler.initialize_dictionary(
            # wavelet_election = wavelet_election,
            # shuffle = shuffle_dictionary
        )

    def encode(self, input_file, output_file):
        """Compress input_file with the given parameters into output_file"""

        image = Image.open(input_file)
        image = image.convert("YCbCr")  # por que esta eleccion en lugar de RGB?
        w, h = image.size
        depth = Utility.mode_to_bpp(image.mode) // 8
        self.image_rawsize = w * h * depth  # no usado hasta ahora
        self.processed_blocks = 0
        self.non_zero_coefs = 0

        with RawFile(output_file, "wb") as file:

            file.write_header(
                self.header_format,
                self.magic_number,
                self.fif_version,
                w,
                h,
                depth,
                0,  # A_id (to be removed)
                0,  # basis_index (to be removed)
                self.min_n,
                self.max_n,
            )

            image_data = np.array(image.getdata()).reshape(h, w, depth)

            n = self.max_n
            x_list = []
            for k in range(depth):
                channel_processed_blocks, x_list = self.omp_handler.omp_encode(
                    x_list=x_list,
                    image_data=image_data[:, :, k],
                    max_error=self.max_error,
                    block_size=n,
                    k=k,
                )
                self.processed_blocks += channel_processed_blocks

            for n, i, j, k, x in x_list:
                file.write("H", i)
                file.write("H", j)
                file.write("B", k)
                file.write("B", n)

                x = Utility.quantize(x, self.v_format_precision).tolist()
                x_norm_0 = np.linalg.norm(x, 0)
                self.non_zero_coefs += x_norm_0

                file.write_vector(x, x_norm_0, self.v_format_precision)
                # print("i, j, k, n: ", i, j, k, n)

            file.write("I", self.processed_blocks)

            bytes_written = file.tell()
        print(f"processed_blocks: {self.processed_blocks}")
        print(f"bytes_written (without DEFLATE): {bytes_written}")

        if self.apply_deflate == True:
            ## apply DEFLATE compression
            with open(output_file, "rb") as file:
                compress = zlib.compressobj(
                    9, zlib.DEFLATED, 15, 9, zlib.Z_DEFAULT_STRATEGY
                )  # ver estos parametros
                zdata = compress.compress(file.read())
                zdata += compress.flush()

            with open(output_file, "wb") as file:
                file.write(zdata)
                print(f"DEFLATE applied. Bytes to write: {file.tell()}")

        print("File saved.")

    def decode(self, input_file, output_file):
        """Decompress input_file into output_file"""

        if self.apply_deflate == True:
            # se descomprime el archivo al que se le aplicó DEFLATE
            with open(input_file, "rb") as file:
                decompress = zlib.decompressobj(
                    15
                )  # 15 is the window size (must match the compress settings)
                decompressed_data = decompress.decompress(file.read())
                decompressed_data += decompress.flush()

            new_file = input_file + "_after_deflate"
            with open(new_file, "wb") as file:
                file.write(decompressed_data)
            input_file = new_file

        with RawFile(input_file, "rb") as file:
            file.seek(-4, 2)
            processed_blocks = file.read("I")
            file.seek(0)
            # print("processed_blocks:",processed_blocks)
            # No es necesario ya que la clase guarda estos parámetros
            # A_id, basis_index = 0, 0 (to be removed)
            magic_number_read, version, w, h, depth, A_id, basis_index, min_n, max_n = (
                file.read(self.header_format)
            )
            # print(magic_number_read, version, w, h, depth, A_id, basis_index, min_n, max_n)

            if magic_number_read != self.magic_number.decode("utf-8"):
                raise Exception(
                    f"Invalid image format: Wrong magic number '{magic_number_read}'"
                )

            if version != self.fif_version:
                raise Exception(
                    f"Invalid codec version: {version}. Expected: {self.fif_version}"
                )

            image_data = np.zeros((h, w, depth), dtype=np.float32)
            image_data = self.omp_handler.omp_decode(
                file, image_data, self.v_format_precision, processed_blocks
            )

            if depth == 1:
                image_data[:, :, 1] = image_data[:, :, 0]
                image_data[:, :, 2] = image_data[:, :, 0]

            image_data = Utility.ycbcr_to_rgb(image_data)
            image = Image.fromarray(image_data.astype("uint8"))
            image.save(output_file)
            print("Output file saved to: " + output_file)



from setuptools import setup, find_packages

setup(
    name="aqmp-compressor",
    version="0.1.5",
    description="A Python package for compressing and decompressing images using OMP and wavelet transforms.",
    author="Emmanuel A. Tassone",
    author_email="emmanueltassone@gmail.com",
    url="https://github.com/Emmatassone/AQMP",
    packages=find_packages(),
    install_requires=[
        "numpy",
        "scikit-learn",
        "Pillow",
        "PyWavelets",
        "anytree"
    ],
    entry_points={
        "console_scripts": [
            "aqmp=aqmp.main:main",
        ],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='<=3.11, >=3.9',
)